#-----------------------------------------------------------------------------
# Copyright 2008, Victor Liu
#
# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------

class BezPatchEdit

    def BezPatchEdit.select_tool
        Sketchup.active_model.select_tool BezPatchEdit.new
    end
    
    def initialize
        @ip1 = Sketchup::InputPoint.new
        @ip2 = Sketchup::InputPoint.new
    end
    
    def reset
        @drawn = false
        @dragging = false
        @group = nil
        @picked_index = nil
        @picked_line = nil
        @lit_pt = nil
        @axis_lock = nil
        @bez = nil
        @bezLines = nil
        @shift_down_time = Time.now
        @ip1.clear
        @ip2.clear
    end

    def _show_vcb
        Sketchup::set_status_text "Bezier steps [U,V]", SB_VCB_LABEL
        Sketchup::set_status_text "#{@uSteps}, #{@vSteps}", SB_VCB_VALUE
    end
    
    def activate
        self.reset
        
        @group = BezPatch.selected_patch
        if (! @group)
            Sketchup.active_model.select_tool nil
            return
        end
        
        # Get bezier patch attributes.
        @uOrder = @group.get_attribute('BezPatch', 'uOrder')
        @vOrder = @group.get_attribute('BezPatch', 'vOrder')
        @uSteps = @group.get_attribute('BezPatch', 'uSteps')
        @vSteps = @group.get_attribute('BezPatch', 'vSteps')
        ctrlPts = @group.get_attribute('BezPatch', 'ctrlPts')
        @bezType = @group.get_attribute('BezPatch', 'bezType')
        
        # Recover control points as Point3ds, and transform to group space.
        t = @group.transformation
        @ctrlPts = []
        ctrlPts.each do |a|
            pt = Geom::Point3d.new(a[0], a[1], a[2])
            pt.transform!(t)
            @ctrlPts.push(pt)
        end
        
        # Create BezPatch instance.
        @bez = BezPatch.new({
            'uOrder' => @uOrder,
            'vOrder' => @vOrder,
            'uSteps' => @uSteps,
            'vSteps' => @vSteps,
            'ctrlPts' => @ctrlPts,
            'bezType' => @bezType
        }, @group.transformation)
        
        # Set up control array lines.
        # @uLines run along U direction, and there are (@vOrder+1) of them.
        # @vLines are similarly defined.
        @uLines = []
        @vLines = []
        for j in (0..@vOrder)
            @uLines[j] = []
        end
        for i in (0..@uOrder)
            @vLines[i] = []
        end
        p = 0
        for j in (0..@vOrder)
            for i in (0..@uOrder)
                pt = @ctrlPts[p]
                @uLines[j].push(pt)
                @vLines[i].push(pt)
                p += 1
            end
        end
        
        self._show_vcb
    end

    def deactivate(view)
        view.invalidate
    end

    def commit_changes
        # Replace old bezier patch with the newer one.
        if (@group)
            # Inverse transform control points.
            t = @group.transformation.inverse
            @ctrlPts.each do |pt|
                pt.transform!(t)
            end
            
            @bez.create_patch(@group)
        end
        Sketchup.active_model.select_tool nil
    end

    def onCancel(flag, view)
        self.reset
        Sketchup.active_model.select_tool nil
    end
    
    def resume(view)
        view.invalidate
        self._show_vcb
    end

    def _pick(x, y, view)
        old_lit_pt = @lit_pt
        @picked_line = nil
        
        # Run through all lines of control array, looking for selection
        ph = view.pick_helper x, y
        lines = @uLines + @vLines
        lines.each do |line|
            @picked_index = ph.pick_segment line
            if (@picked_index)
                if (@picked_index < 0)
                    # Picked a segment.  Compute the point closest
                    # to the pick ray.
                    pickray = view.pickray x, y
                    i = -@picked_index
                    segment = [line[i-1], line[i]]
                    result = Geom.closest_points segment, pickray
                    @lit_pt = result[0]
                else
                    # Picked a control point.
                    @lit_pt = line[@picked_index]
                end
                @picked_line = line
                break
            else
                @lit_pt = nil
            end
        end
        
        return old_lit_pt != @lit_pt
    end

    def onMouseMove(flags, x, y, view)
        @x = x
        @y = y
        view.invalidate if not @drawn
    
        if (@dragging && @picked_line && @picked_index)
            if (@axis_lock)
                offset_pt = @axis_lock.offset_point(x, y, view)
            else
                # Move picked selection
                @ip1.pick view, x, y
                return if not @ip1.valid?
                offset_pt = @ip1.position
            end
            
            if (@picked_index >= 0)
                # Moving a control point
                @lit_pt = offset_pt
                @picked_line[@picked_index].set!(
                    @lit_pt.x, @lit_pt.y, @lit_pt.z)
            else
                # Moving a segment between control points
                vec = offset_pt - @lit_pt
                i = -@picked_index
                @picked_line[i-1].offset!(vec)
                @picked_line[i].offset!(vec)
                @lit_pt = offset_pt
            end
            
            # Update bez patch.
            @bez.ctrlPts = @ctrlPts
            @bezLines = BezPatch.lines(@bez.points, @bez.uSteps, @bez.vSteps)
            
            view.invalidate
        else # Nothing picked
            # See if we can select something to move
            view.invalidate if (self._pick(x, y, view))
        end
    end

    def onLButtonDown(flags, x, y, view)
        # Select the segment or control point to move
        if (@axis_lock == nil)
            self._pick x, y, view
        end
        @dragging = true
        view.lock_inference
    end

    def onLButtonUp(flags, x, y, view)
        @dragging = false
        @axis_lock = nil
        if ! @picked_line || ! @picked_index
            view.invalidate
            return
        end
        view.invalidate
    end

    def draw(view)
        return unless @bez
        
        bg = Sketchup.active_model.rendering_options['BackgroundColor']
        nbg = Sketchup::Color.new(bg.red^0xFF, bg.green^0xFF, bg.blue^0xFF)
        if (@bezLines)
            # Display modified patch.
            view.drawing_color = Sketchup::Color.blend(bg, nbg, 0.5)
            for j in (0..@vSteps)
                view.draw(GL_LINE_STRIP, @bezLines[0][j])
            end
            for i in (0..@uSteps)
                view.draw(GL_LINE_STRIP, @bezLines[1][i])
            end
        end
        
        # Display control array.
        view.drawing_color = nbg
        for j in (0..@vOrder)
            view.draw(GL_LINE_STRIP, @uLines[j])
        end
        for i in (0..@uOrder)
            view.draw(GL_LINE_STRIP, @vLines[i])
        end
        view.draw_points(@ctrlPts, 10, 1, 'gray')
        
        if (@lit_pt)
            color = 'yellow'
            if (@axis_lock)
                case @axis_lock.axis
                when X_AXIS
                    color = 'red'
                when Y_AXIS
                    color = 'green'
                when Z_AXIS
                    color = 'blue'
                end
                view.drawing_color = color
                view.line_stipple = '-'
                view.draw_line(@axis_lock.origin, @lit_pt)
                view.line_stipple = 0                
            end
            view.draw_points(@lit_pt, 10, 1, color)
        end
        
        @drawn = true
    end
    
    def _lock_to_axis(axis, view)
        return if (! @picked_line || ! @picked_index)
        
        # Begin tracking displacement in screen coordinates.
        @axis_lock = AxisLock.new(@x, @y, @lit_pt.clone, axis)
        
        # Paint color change of @lit_pt
        view.invalidate
    end
    
    def onKeyDown(key, rpt, flags, view)
        case (key)
        when 27  # ESCAPE
            Sketchup.active_model.select_tool nil
        when CONSTRAIN_MODIFIER_KEY
            @shift_down_time = Time.now
        
            # if we already have an inference lock, then unlock it
            if (view.inference_locked?)
                view.lock_inference
            elsif (@dragging)
                view.lock_inference @ip1, @ip2
            else
                view.lock_inference @ip1
            end
        when VK_LEFT
            _lock_to_axis(Y_AXIS, view)
        when VK_UP
            _lock_to_axis(Z_AXIS, view)
        when VK_RIGHT
            _lock_to_axis(X_AXIS, view)
        end
    end

    def onKeyUp(key, rpt, flags, view)
        if (key == CONSTRAIN_MODIFIER_KEY &&
            view.inference_locked? &&
            (Time.now - @shift_down_time) > 0.5
            )
            view.lock_inference
        end
    end

    def onUserText(text, view)
        begin
            uSteps, vSteps = text.split(',')
            uSteps = uSteps.to_i
            vSteps = vSteps.to_i
        rescue
            # Error parsing the text
            UI.beep
            uSteps = vSteps = nil
        end
        if (!uSteps || !vSteps ||
            uSteps <= @uOrder || vSteps <= @vOrder)
            self._show_vcb
            return
        end
        
        @bez.uSteps = @uSteps = uSteps
        @bez.vSteps = @vSteps = vSteps
        @bezLines = BezPatch.lines(@bez.points, uSteps, vSteps)
        view.invalidate
        
        self._show_vcb
    end

    def getMenu(menu)
        menu.add_item("Done") { commit_changes }
        menu.add_item("Cancel") { Sketchup.active_model.select_tool nil }
    end

    def getExtents
        bb = Geom::BoundingBox.new
        bb.add @ctrlPts
        bb
    end

end
