#-----------------------------------------------------------------------------
# Copyright 2008, Victor Liu
#
# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------

class BezPatchTool

    def BezPatchTool.select_tool
        Sketchup.active_model.select_tool BezPatchTool.new
    end

    def BezPatchTool.select_grid_tool
        Sketchup.active_model.select_tool BezPatchTool.new('Grid')
    end

    @@uOrder = 2
    @@vOrder = 2
    @@uSteps = BezPatch::DFLT_NR_STEPS
    @@vSteps = BezPatch::DFLT_NR_STEPS

    def initialize(bezType = 'Patch')
        @bezType = bezType
        @ip1 = Sketchup::InputPoint.new
        @ip2 = Sketchup::InputPoint.new
        reset
    end

    def reset
        @pts = []
        @state = 0
        @ip2.clear
        @drawn = false
        Sketchup::set_status_text "Click for start point"
        @shift_down_time = Time.now
    end

    def _show_vcb
        Sketchup::set_status_text "Bezier order [U,V]", SB_VCB_LABEL
        Sketchup::set_status_text "#{@@uOrder}, #{@@vOrder}", SB_VCB_VALUE
    end
    
    def _get_bez_params
        # Get input params.
        prompts = [ 'U Order', 'V Order', 'U Steps', 'V Steps' ]
        dflts = [ @@uOrder, @@vOrder, @@uSteps, @@vSteps ]
        results = inputbox(prompts, dflts, "Bezier #{@bezType} Params")
        return if not results
        
        @@uOrder = results[0]
        @@vOrder = results[1]
        @@uSteps = results[2]
        @@vSteps = results[3]
        
        return results
    end

    def activate
        params = _get_bez_params
        if (!params)
            Sketchup.active_model.select_tool nil
            return
        end

        self.reset
        @drawn = false
        self._show_vcb
    end

    def deactivate(view)
        view.invalidate if @drawn
    end

    def _set_current_point(x, y, view)
        if (! @ip1.pick(view, x, y, @ip2))
            return false
        end
        need_draw = true
    
        # Set the tooltip that will be displayed
        view.tooltip = @ip1.tooltip
        
        # Compute points
        case @state
        when 0
            @pts[0] = @ip1.position
            @pts[4] = @pts[0]
            need_draw = @ip1.display? || @drawn
        when 1
            @pts[1] = @ip1.position
        when 2
            pt1 = @ip1.position
            pt2 = pt1.project_to_line @pts
            vec = pt1 - pt2
            @pts[2] = @pts[1] + vec
            @pts[3] = @pts[0] + vec
        end

        view.invalidate if need_draw
    end

    def _increment_state
        @state += 1
        case @state
        when 1
            @ip2.copy! @ip1
            Sketchup::set_status_text "Click for second point"
        when 2
            @ip2.clear
            @pts[2] = @pts[1]
            @pts[3] = @pts[0]
            Sketchup::set_status_text "Click for third point"
        when 3
            self._create_bez_patch
        end
    end

    def _create_bez_patch
        data = {}
        data['uOrder'] = @@uOrder
        data['vOrder'] = @@vOrder
        data['bezType'] = @bezType
        data['uSteps'] = @@uSteps if @@uSteps
        data['vSteps'] = @@vSteps if @@vSteps
        
        ## Determine matrix of control points.  Order is along U, then V.
        du = (@pts[1] - @pts[0]).length / @@uOrder
        dv = (@pts[3] - @pts[0]).length / @@vOrder
        ctrlPts = []
        for j in (0..@@vOrder) do
            for i in (0..@@uOrder) do
                ctrlPts.push(Geom::Point3d.new(i*du, j*dv, 0.0))
            end
        end
        data['ctrlPts'] = ctrlPts
        
        # Transformation.
        xaxis = @pts[1] - @pts[0]
        yaxis = @pts[3] - @pts[0]
        zaxis = xaxis.cross(yaxis)
        t = Geom::Transformation.axes(@pts[0], xaxis, yaxis, zaxis)
        
        bez = BezPatch.new(data, t)
        bez.create_patch
    end

    def onMouseMove(flags, x, y, view)
        self._set_current_point(x, y, view)
    end

    def onLButtonDown(flags, x, y, view)
        self._set_current_point(x, y, view)
        self._increment_state
        view.lock_inference
    end

    def onLButtonUp(flags, x, y, view)
        Sketchup.active_model.select_tool nil if @state == 3
    end

    def onCancel(flag, view)
        view.invalidate if @drawn
        self.reset
    end

    def getExtents
        bb = Geom::BoundingBox.new
        case @state
        when 0
            # We are getting the first point
            if (@ip1.valid? && @ip1.display?)
                bb.add @ip1.position
            end
        when 1
            bb.add @pts[0]
            bb.add @pts[1]
        when 2
            bb.add @pts
        end
        bb
    end

    # This is called when the user types a value into the VCB
    def onUserText(text, view)
        begin
            uOrder, vOrder = text.split(',')
            uOrder = uOrder.to_i
            vOrder = vOrder.to_i
        rescue
            # Error parsing the text
            UI.beep
            uOrder = vOrder = nil
        end
        if (!uOrder || !vOrder)
            self._show_vcb
            return
        end
        
        @@uOrder = uOrder
        @@vOrder = vOrder
        self._show_vcb
    end

    def onKeyDown(key, rpt, flags, view)
        if (key == CONSTRAIN_MODIFIER_KEY)
            @shift_down_time = Time.now
        
            # if we already have an inference lock, then unlock it
            if (view.inference_locked?)
                view.lock_inference
            elsif (@state == 1)
                view.lock_inference @ip1, @ip2
            else
                view.lock_inference @ip1
            end
        end
    end

    def onKeyUp(key, rpt, flags, view)
        if (key == CONSTRAIN_MODIFIER_KEY &&
            view.inference_locked? &&
            (Time.now - @shift_down_time) > 0.5
            )
            view.lock_inference
        end
    end

    def _draw_grid_lines(view, ptA0, ptA1, ptB0, ptB1, nLines)
        t = 1.0/nLines
        d0 = Geom::Point3d.linear_combination(1.0-t, ptA0, t, ptB0)
        d1 = Geom::Point3d.linear_combination(1.0-t, ptA1, t, ptB1)
        v0 = ptA0.vector_to d0
        v1 = ptA1.vector_to d1
        
        view.draw(GL_LINE_STRIP, ptA0, ptB0)
        for i in (0...nLines-1)
            view.draw(GL_LINE_STRIP, d0, d1)
            d0.offset!(v0)
            d1.offset!(v1)
        end
        view.draw(GL_LINE_STRIP, ptA1, ptB1)
    end

    def _draw_grid(view)
        # Draw lines showing initial array of control points.
        col = Sketchup.active_model.rendering_options['BackgroundColor']
        col = Sketchup::Color.new(col.red^0xFF, col.green^0xFF, col.blue^0xFF)
        view.drawing_color = col
        _draw_grid_lines(view, @pts[0], @pts[1], @pts[3], @pts[2], @@uOrder)
        _draw_grid_lines(view, @pts[0], @pts[3], @pts[1], @pts[2], @@vOrder)
        return true
    end
    
    def draw(view)
        @drawn = false
    
        # Show the current input point
        if (@ip1.valid? && @ip1.display?)
            @ip1.draw(view)
            @drawn = true
        end

        # show the rectangle
        if (@state == 1)
            # just draw a line from the start to the end point
            view.set_color_from_line(@ip2, @ip1)
            view.draw(GL_LINE_STRIP, @pts[0], @pts[1])
            @drawn = true
        elsif (@state > 1)
            @drawn = _draw_grid(view)
        end
    end

end # of class BezPatchTool

